from django.dispatch import Signal


pre_workdir_copy_to_remote = Signal(
    providing_args=["deploy_revision", "remote_revision", "workdir"]
)
post_workdir_copy_to_remote = Signal(
    providing_args=["deploy_revision", "remote_revision", "workdir"]
)


pre_remote_run_commands = Signal()
post_remote_run_commands = Signal()


pre_remote_activate_revision = Signal()
post_remote_activate_revision = Signal()


pre_remote_reload = Signal()
post_remote_reload = Signal()


pre_remote_init_virtualenv = Signal()
post_remote_init_virtualenv = Signal()



from __future__ import with_statement

import os
import shutil
import glob

from fabric.api import run, local, cd, settings, prefix, hide
from fabric.operations import require
from fabric.state import env
from fabric.contrib import files as fabric_files

from coat import utils as coat_utils
from coat.django.utils import find_settings, find_manage
from coat.django import signals


def copy_revision_to_remote(workdir, remote_revision, deploy_revision):
    # test remote for the revision and skip re-copying already
    # existing revisions
    remote_versions_dir = "%s/%s" % (
        env.base_dir, env.django_settings.versions_dir
    )

    if not fabric_files.exists(remote_versions_dir):
        run("mkdir -p %s" % remote_versions_dir)

    if fabric_files.exists("%s/%s" % (remote_versions_dir,
                                      deploy_revision)):
        return

    # copy all of local onto remote using rsync, use hard links to save
    # space for unmodified files
    rsync_cmd = ("%s/* %s@%%s:%s/%s/" %
                 (workdir, env.user,
                  remote_versions_dir,
                  deploy_revision))

    signals.pre_workdir_copy_to_remote.send(
        sender=copy_revision_to_remote,
        deploy_revision=deploy_revision,
        remote_revision=remote_revision,
        workdir=workdir,
    )

    if remote_revision:
        with cd("%s/%s" % remote_versions_dir):
            run("rsync -a --link-dest=../%(cur)s/ %(cur)s/ %(new)s" %
                {'cur': remote_revision, 'new': deploy_revision})

        rsync_cmd = "-a -c --delete " + rsync_cmd
    else:
        rsync_cmd = "-a " + rsync_cmd

    for host in env.hosts:
        local("rsync %s" % (rsync_cmd % host))

    signals.post_workdir_copy_to_remote.send(
        sender=copy_revision_to_remote,
        deploy_revision=deploy_revision,
        remote_revision=remote_revision,
        workdir=workdir,
    )


def remote_activate_revision(workdir, remote_revision, deploy_revision):
    remote_virtualenv_dir = "%s/%s" % (
        env.base_dir, env.virtualenv_settings.env_dir
    )

    remote_versions_dir = "%s/%s" % (
        env.base_dir, env.django_settings.versions_dir
    )

    signals.pre_remote_run_commands.send(
        sender=remote_activate_revision,
    )

    # initialize virtualenv if non-existant
    if not fabric_files.exists(remote_virtualenv_dir):
        signals.pre_remote_init_virtualenv.send(
            sender=remote_activate_revision
        )

        for command in env.virtualenv_settings.init_commands:
            run(command)

        signals.post_remote_init_virtualenv.send(
            sender=remote_activate_revision
        )

    # find relative path to manage.py
    django_manage_path = find_manage(workdir).replace(workdir, "")[1:]

    with cd("%s/%s" % (remote_versions_dir, deploy_revision)):
        with prefix("source %s/bin/activate" % remote_virtualenv_dir):
            for command in env.virtualenv_settings.commands:
                run(command)

        for command in env.django_settings.management_commands:
            run("%s/bin/python %s %s" % (
                remote_virtualenv_dir, django_manage_path, command
            ))

    signals.post_remote_run_commands.send(
        sender=remote_activate_revision,
    )

    signals.pre_remote_activate_revision.send(
        sender=remote_activate_revision,
    )

    with cd(remote_versions_dir):
        with settings(hide("warnings", "stderr"), warn_only=True):
            run("test -L current && rm current")

        run("ln -s %s current" % deploy_revision)

    signals.post_remote_activate_revision.send(
        sender=remote_activate_revision,
    )


def remote_reload():
    signals.pre_remote_reload.send(sender=remote_reload)

    if env.django_settings.wsgi_file:
        run("touch %s/%s" % (env.base_dir, env.django_settings.wsgi_file))

    signals.post_remote_reload.send(sender=remote_reload)


def workdir_django_prepare(workdir):
    if 'settings_file' in env.django_settings:
        django_basedir = os.path.dirname(find_settings(workdir))

        shutil.copy(
            os.path.join(django_basedir, env.django_settings.settings_file),
            os.path.join(django_basedir, "localsettings.py")
        )

        map(os.unlink, glob.glob("%s/localsettings_*.py" % django_basedir))


def deploy(revision="HEAD"):
    require("django_settings", "base_dir", "virtualenv_settings", provided_by=("env_test", "env_live"))

    env.django_settings.validate_or_abort()
    env.virtualenv_settings.validate_or_abort()

    env.remote_revision = coat_utils.remote_resolve_current_revision()
    env.deploy_revision = coat_utils.local_resolve_revision(revision)

    env.deploy_workdir = coat_utils.workdir_prepare_checkout(
        revision, folders=("django", )
    )

    workdir_django_prepare(env.deploy_workdir)

    copy_revision_to_remote(
        env.deploy_workdir, env.remote_revision, env.deploy_revision
    )

    remote_activate_revision(
        env.deploy_workdir, env.remote_revision, env.deploy_revision
    )

    remote_reload()



import os
import atexit
import shutil
import tempfile

from fabric.state import env
from fabric.api import lcd, local, hide, settings, cd, run
from fabric.contrib import files as fabric_files
from coat import signals


def get_project_root_directory():
    """
    Returns the dirname of the fabfile.
    """
    return os.path.dirname(os.path.abspath(env.real_fabfile))


def workdir_prepare_checkout(revision, folders):
    """
    Returns an absolute path to a directory containing a git checkout of the
    given treeish revision.

    Registers a hook with atexit to delete the returned directory.

    Dispatches signals to pre_workdir_checkout and post_workdir_checkout.
    """
    signals.pre_workdir_prepare_checkout.send(
        sender=workdir_prepare_checkout,
        revision=revision,
    )

    workdir = tempfile.mkdtemp()

    atexit.register(shutil.rmtree, workdir, True)

    with lcd(get_project_root_directory()):
        local("git archive %s %s | tar -x -C %s" %
              (revision, " ".join(folders), workdir))

        signals.post_workdir_prepare_checkout.send(
            sender=workdir_prepare_checkout,
            revision=revision,
            workdir=workdir,
        )

    return workdir


def local_resolve_revision(revision):
    """
    Returns a partial sha1 from a treeish revision.
    """
    with lcd(get_project_root_directory()):
        with hide("running"):
            revision = local("git log -1 --format=%%h %s" % revision,
                             capture=True)

    return revision


def remote_resolve_current_revision():
    """
    Returns the locally resolved currently active remote revision.
    """
    revision = None

    with settings(hide("stderr", "running", "stdout", "warnings"),
                  warn_only=True):

        # make sure version_dir exists
        if not fabric_files.exists(env.django_settings.versions_dir):
            run("mkdir -p %(versions_dir)s" % env.django_settings)

        # try to resolve current symlink
        with cd(env.django_settings.versions_dir):
            if fabric_files.exists("current"):
                revision = run("readlink current")

                if not revision.failed:
                    revision = revision.rstrip("/")

    if revision:
        return local_resolve_revision(revision)
